"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Browser = void 0;
const os = require("os");
const uniqueFilename = require("unique-filename");
const puppeteer = require("puppeteer");
const chokidar = require("chokidar");
const path = require("path");
const fs = require("fs");
class Browser {
    constructor(config, log, metrics) {
        this.config = config;
        this.log = log;
        this.metrics = metrics;
        this.logError = (err) => {
            this.log.error('Browser page crashed', 'error', err.toString());
        };
        this.logPageError = (err) => {
            this.log.error('Browser uncaught exception', 'error', err.toString());
        };
        this.logConsoleMessage = (msg) => {
            const msgType = msg.type();
            if (!this.config.verboseLogging && msgType !== 'error') {
                return;
            }
            const loc = msg.location();
            if (msgType === 'error') {
                this.log.error('Browser console error', 'msg', msg.text(), 'url', loc.url, 'line', loc.lineNumber, 'column', loc.columnNumber);
                return;
            }
            this.log.debug(`Browser console ${msgType}`, 'msg', msg.text(), 'url', loc.url, 'line', loc.lineNumber, 'column', loc.columnNumber);
        };
        this.logRequest = (req) => {
            this.log.debug('Browser request', 'url', req.url(), 'method', req.method());
        };
        this.logRequestFailed = (req) => {
            this.log.error('Browser request failed', 'url', req.url(), 'method', req.method(), 'failure', req.failure().errorText);
        };
        this.logRequestFinished = (req) => {
            this.log.debug('Browser request finished', 'url', req.url(), 'method', req.method());
        };
        this.logPageClosed = () => {
            this.log.debug('Browser page closed');
        };
        this.log.debug('Browser initialized', 'config', this.config);
    }
    getBrowserVersion() {
        return __awaiter(this, void 0, void 0, function* () {
            let browser;
            try {
                const launcherOptions = this.getLauncherOptions({});
                browser = yield puppeteer.launch(launcherOptions);
                return yield browser.version();
            }
            finally {
                if (browser) {
                    const pages = yield browser.pages();
                    yield Promise.all(pages.map((page) => page.close()));
                    yield browser.close();
                }
            }
        });
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () { });
    }
    validateRenderOptions(options) {
        if (options.url.startsWith(`socket://`)) {
            // Puppeteer doesn't support socket:// URLs
            throw new Error(`Image rendering in socket mode is not supported`);
        }
        options.headers = options.headers || {};
        const headers = {};
        if (options.headers['Accept-Language']) {
            headers['Accept-Language'] = options.headers['Accept-Language'];
        }
        else if (this.config.acceptLanguage) {
            headers['Accept-Language'] = this.config.acceptLanguage;
        }
        options.headers = headers;
        options.timeout = parseInt(options.timeout, 10) || 30;
    }
    validateImageOptions(options) {
        this.validateRenderOptions(options);
        options.width = parseInt(options.width, 10) || this.config.width;
        options.height = parseInt(options.height, 10) || this.config.height;
        if (options.width < 10) {
            options.width = this.config.width;
        }
        if (options.width > this.config.maxWidth) {
            options.width = this.config.maxWidth;
        }
        if (options.height < 10) {
            options.height = this.config.height;
        }
        if (options.height > this.config.maxHeight) {
            options.height = this.config.maxHeight;
        }
        options.deviceScaleFactor = parseFloat((options.deviceScaleFactor || '1')) || 1;
        if (options.deviceScaleFactor > this.config.maxDeviceScaleFactor) {
            options.deviceScaleFactor = this.config.deviceScaleFactor;
        }
    }
    getLauncherOptions(options) {
        const env = Object.assign({}, process.env);
        // set env timezone
        env.TZ = options.timezone || this.config.timezone;
        const launcherOptions = {
            env: env,
            ignoreHTTPSErrors: this.config.ignoresHttpsErrors,
            dumpio: this.config.dumpio,
            args: this.config.args,
        };
        if (this.config.chromeBin) {
            launcherOptions.executablePath = this.config.chromeBin;
        }
        return launcherOptions;
    }
    setTimezone(page, options) {
        return __awaiter(this, void 0, void 0, function* () {
            const timezone = options.timezone || this.config.timezone;
            if (timezone) {
                yield page.emulateTimezone(timezone);
            }
        });
    }
    preparePage(page, options) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.config.verboseLogging) {
                this.log.debug('Setting cookie for page', 'renderKey', options.renderKey, 'domain', options.domain);
            }
            yield page.setCookie({
                name: 'renderKey',
                value: options.renderKey,
                domain: options.domain,
            });
            if (options.headers && Object.keys(options.headers).length > 0) {
                this.log.debug(`Setting extra HTTP headers for page`, 'headers', options.headers);
                yield page.setExtraHTTPHeaders(options.headers);
            }
        });
    }
    render(options) {
        return __awaiter(this, void 0, void 0, function* () {
            let browser;
            let page;
            try {
                browser = yield this.withTimingMetrics(() => {
                    this.validateImageOptions(options);
                    const launcherOptions = this.getLauncherOptions(options);
                    return puppeteer.launch(launcherOptions);
                }, 'launch');
                page = yield this.withTimingMetrics(() => {
                    return browser.newPage();
                }, 'newPage');
                this.addPageListeners(page);
                return yield this.takeScreenshot(page, options);
            }
            finally {
                if (page) {
                    this.removePageListeners(page);
                    yield page.close();
                }
                if (browser) {
                    yield browser.close();
                }
            }
        });
    }
    takeScreenshot(page, options) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.withTimingMetrics(() => __awaiter(this, void 0, void 0, function* () {
                if (this.config.verboseLogging) {
                    this.log.debug('Setting viewport for page', 'width', options.width.toString(), 'height', options.height.toString(), 'deviceScaleFactor', options.deviceScaleFactor.toString());
                }
                yield page.setViewport({
                    width: options.width,
                    height: options.height,
                    deviceScaleFactor: options.deviceScaleFactor,
                });
                yield this.preparePage(page, options);
                yield this.setTimezone(page, options);
                if (this.config.verboseLogging) {
                    this.log.debug('Moving mouse on page', 'x', options.width, 'y', options.height);
                }
                return page.mouse.move(options.width, options.height);
            }), 'prepare');
            yield this.withTimingMetrics(() => {
                if (this.config.verboseLogging) {
                    this.log.debug('Navigating and waiting for all network requests to finish', 'url', options.url);
                }
                return page.goto(options.url, { waitUntil: 'networkidle0', timeout: options.timeout * 1000 });
            }, 'navigate');
            yield this.withTimingMetrics(() => {
                if (this.config.verboseLogging) {
                    this.log.debug('Waiting for dashboard/panel to load', 'timeout', `${options.timeout}s`);
                }
                return page.waitForFunction(() => {
                    const panelCount = document.querySelectorAll('.panel').length || document.querySelectorAll('.panel-container').length;
                    return window.panelsRendered >= panelCount;
                }, {
                    timeout: options.timeout * 1000,
                });
            }, 'panelsRendered');
            if (!options.filePath) {
                options.filePath = uniqueFilename(os.tmpdir()) + '.png';
            }
            if (this.config.verboseLogging) {
                this.log.debug('Taking screenshot', 'filePath', options.filePath);
            }
            yield this.withTimingMetrics(() => {
                return page.screenshot({ path: options.filePath });
            }, 'screenshot');
            return { filePath: options.filePath };
        });
    }
    renderCSV(options) {
        return __awaiter(this, void 0, void 0, function* () {
            let browser;
            let page;
            try {
                this.validateRenderOptions(options);
                const launcherOptions = this.getLauncherOptions(options);
                browser = yield puppeteer.launch(launcherOptions);
                page = yield browser.newPage();
                this.addPageListeners(page);
                return yield this.exportCSV(page, options);
            }
            finally {
                if (page) {
                    this.removePageListeners(page);
                    yield page.close();
                }
                if (browser) {
                    yield browser.close();
                }
            }
        });
    }
    exportCSV(page, options) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.preparePage(page, options);
            yield this.setTimezone(page, options);
            const downloadPath = uniqueFilename(os.tmpdir());
            fs.mkdirSync(downloadPath);
            const watcher = chokidar.watch(downloadPath);
            let downloadFilePath = '';
            watcher.on('add', (file) => {
                if (!file.endsWith('.crdownload')) {
                    downloadFilePath = file;
                }
            });
            yield page._client.send('Page.setDownloadBehavior', { behavior: 'allow', downloadPath: downloadPath });
            if (this.config.verboseLogging) {
                this.log.debug('Navigating and waiting for all network requests to finish', 'url', options.url);
            }
            yield page.goto(options.url, { waitUntil: 'networkidle0', timeout: options.timeout * 1000 });
            if (this.config.verboseLogging) {
                this.log.debug('Waiting for download to end');
            }
            const startDate = Date.now();
            while (Date.now() - startDate <= options.timeout * 1000) {
                if (downloadFilePath !== '') {
                    break;
                }
                yield new Promise((resolve) => setTimeout(resolve, 500));
            }
            if (downloadFilePath === '') {
                throw new Error(`Timeout exceeded while waiting for download to end`);
            }
            yield watcher.close();
            let filePath = downloadFilePath;
            if (options.filePath) {
                fs.copyFileSync(downloadFilePath, options.filePath);
                fs.unlinkSync(downloadFilePath);
                fs.rmdirSync(path.dirname(downloadFilePath));
                filePath = options.filePath;
            }
            return { filePath, fileName: path.basename(downloadFilePath) };
        });
    }
    withTimingMetrics(callback, step) {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.config.timingMetrics) {
                const endTimer = this.metrics.durationHistogram.startTimer({ step });
                const res = yield callback();
                endTimer();
                return res;
            }
            else {
                return callback();
            }
        });
    }
    addPageListeners(page) {
        page.on('error', this.logError);
        page.on('pageerror', this.logPageError);
        page.on('requestfailed', this.logRequestFailed);
        page.on('console', this.logConsoleMessage);
        if (this.config.verboseLogging) {
            page.on('request', this.logRequest);
            page.on('requestfinished', this.logRequestFinished);
            page.on('close', this.logPageClosed);
        }
    }
    removePageListeners(page) {
        page.removeListener('error', this.logError);
        page.removeListener('pageerror', this.logPageError);
        page.removeListener('requestfailed', this.logRequestFailed);
        page.removeListener('console', this.logConsoleMessage);
        if (this.config.verboseLogging) {
            page.removeListener('request', this.logRequest);
            page.removeListener('requestfinished', this.logRequestFinished);
            page.removeListener('close', this.logPageClosed);
        }
    }
}
exports.Browser = Browser;
//# sourceMappingURL=browser.js.map