"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.HttpServer = void 0;
const fs = require("fs");
const path = require("path");
const express = require("express");
const boom = require("@hapi/boom");
const morgan = require("morgan");
const promClient = require("prom-client");
const browser_1 = require("../browser");
const metrics_middleware_1 = require("./metrics_middleware");
class HttpServer {
    constructor(config, log) {
        this.config = config;
        this.log = log;
        this.render = (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            if (!req.query.url) {
                throw boom.badRequest('Missing url parameter');
            }
            const headers = {};
            if (req.headers['Accept-Language']) {
                headers['Accept-Language'] = req.headers['Accept-Language'].join(';');
            }
            const options = {
                url: req.query.url,
                width: req.query.width,
                height: req.query.height,
                filePath: req.query.filePath,
                timeout: req.query.timeout,
                renderKey: req.query.renderKey,
                domain: req.query.domain,
                timezone: req.query.timezone,
                encoding: req.query.encoding,
                deviceScaleFactor: req.query.deviceScaleFactor,
                headers: headers,
            };
            this.log.debug('Render request received', 'url', options.url);
            req.on('close', (err) => {
                this.log.debug('Connection closed', 'url', options.url, 'error', err);
            });
            const result = yield this.browser.render(options);
            res.sendFile(result.filePath, (err) => {
                if (err) {
                    next(err);
                }
                else {
                    try {
                        this.log.debug('Deleting temporary file', 'file', result.filePath);
                        fs.unlinkSync(result.filePath);
                    }
                    catch (e) {
                        this.log.error('Failed to delete temporary file', 'file', result.filePath);
                    }
                }
            });
        });
        this.renderCSV = (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            if (!req.query.url) {
                throw boom.badRequest('Missing url parameter');
            }
            const headers = {};
            if (req.headers['Accept-Language']) {
                headers['Accept-Language'] = req.headers['Accept-Language'].join(';');
            }
            const options = {
                url: req.query.url,
                filePath: req.query.filePath,
                timeout: req.query.timeout,
                renderKey: req.query.renderKey,
                domain: req.query.domain,
                timezone: req.query.timezone,
                encoding: req.query.encoding,
                headers: headers,
            };
            this.log.debug('Render request received', 'url', options.url);
            req.on('close', (err) => {
                this.log.debug('Connection closed', 'url', options.url, 'error', err);
            });
            const result = yield this.browser.renderCSV(options);
            if (result.fileName) {
                res.setHeader('Content-Disposition', `attachment; filename="${result.fileName}"`);
            }
            res.sendFile(result.filePath, (err) => {
                if (err) {
                    next(err);
                }
                else {
                    try {
                        this.log.debug('Deleting temporary file', 'file', result.filePath);
                        fs.unlinkSync(result.filePath);
                        if (!options.filePath) {
                            fs.rmdirSync(path.dirname(result.filePath));
                        }
                    }
                    catch (e) {
                        this.log.error('Failed to delete temporary file', 'file', result.filePath);
                    }
                }
            });
        });
    }
    start() {
        return __awaiter(this, void 0, void 0, function* () {
            this.app = express();
            this.app.use(morgan('combined', {
                skip: (req, res) => {
                    return res.statusCode >= 400;
                },
                stream: this.log.debugWriter,
            }));
            this.app.use(morgan('combined', {
                skip: (req, res) => {
                    return res.statusCode < 400;
                },
                stream: this.log.errorWriter,
            }));
            if (this.config.service.metrics.enabled) {
                (0, metrics_middleware_1.setupHttpServerMetrics)(this.app, this.config.service.metrics, this.log);
            }
            this.app.get('/', (req, res) => {
                res.send('Grafana Image Renderer');
            });
            this.app.get('/render/version', (req, res) => {
                const pluginInfo = require('../../plugin.json');
                res.send({ version: pluginInfo.info.version });
            });
            this.app.get('/render', asyncMiddleware(this.render));
            this.app.get('/render/csv', asyncMiddleware(this.renderCSV));
            this.app.use((err, req, res, next) => {
                if (err.stack) {
                    this.log.error('Request failed', 'url', req.url, 'stack', err.stack);
                }
                else {
                    this.log.error('Request failed', 'url', req.url, 'error', err);
                }
                return res.status(err.output.statusCode).json(err.output.payload);
            });
            if (this.config.service.host) {
                const server = this.app.listen(this.config.service.port, this.config.service.host, () => {
                    const info = server.address();
                    this.log.info(`HTTP Server started, listening at http://${this.config.service.host}:${info.port}`);
                });
            }
            else {
                const server = this.app.listen(this.config.service.port, () => {
                    const info = server.address();
                    this.log.info(`HTTP Server started, listening at http://localhost:${info.port}`);
                });
            }
            const metrics = {
                durationHistogram: new promClient.Histogram({
                    name: 'grafana_image_renderer_step_duration_seconds',
                    help: 'duration histogram of browser steps for rendering an image labeled with: step',
                    labelNames: ['step'],
                    buckets: [0.1, 0.3, 0.5, 1, 3, 5, 10, 20, 30],
                }),
            };
            this.browser = (0, browser_1.createBrowser)(this.config.rendering, this.log, metrics);
            if (this.config.service.metrics.enabled) {
                const browserInfo = new promClient.Gauge({
                    name: 'grafana_image_renderer_browser_info',
                    help: "A metric with a constant '1 value labeled by version of the browser in use",
                    labelNames: ['version'],
                });
                try {
                    const browserVersion = yield this.browser.getBrowserVersion();
                    browserInfo.labels(browserVersion).set(1);
                }
                catch (_a) {
                    this.log.error('Failed to get browser version');
                    browserInfo.labels('unknown').set(0);
                }
            }
            yield this.browser.start();
        });
    }
}
exports.HttpServer = HttpServer;
// wrapper for our async route handlers
// probably you want to move it to a new file
const asyncMiddleware = (fn) => (req, res, next) => {
    Promise.resolve(fn(req, res, next)).catch((err) => {
        if (!err.isBoom) {
            return next(boom.badImplementation(err));
        }
        next(err);
    });
};
//# sourceMappingURL=http-server.js.map